import parseurl from 'parseurl'
import * as proxy from '../proxy.js'
import * as http from '../http.js'
import finalhandler from 'finalhandler'
import createError, { HttpError } from 'http-errors'
import logger from '../logger.js'

function getFSPath (path) {
  const index = path.indexOf('node_modules')
  if (index < 0) return path
  return path.substr(0, 32)
}

function proxyRequest (req, res, { hostname, protocol, port, secure }) {
  return proxy.web(req, res, {
    hostname,
    protocol,
    port,
    headers: { 'proxied-by-ox-proxy': true },
    secure,
    onRes (req, res, proxyRes) {
      if (proxyRes.statusCode >= 400) {
        throw createError(proxyRes.statusCode, proxyRes.message || '')
      }
      return http.pipe(proxyRes, res)
    }
  })
}

export default function ({ frontends }) {
  const cache = new Map()

  return async (req, res, next) => {
    if (req.headers['proxied-by-ox-proxy']) return next()

    const { path } = parseurl(req)

    // three stage system
    // 1. if we have a known path, we can directly request the file from that system
    // 2. check, if a path is configured for frontends and try to find the file there
    // 3. otherwise, request them in order

    const isFsFile = path.startsWith('/@fs')
    const pathToHandler = isFsFile ? getFSPath(path) : path

    // stage 1
    if (cache.has(pathToHandler)) {
      logger.debug(`Found a handler for ${path} in cache`)
      try {
        const { target: { hostname, protocol, port }, secure } = cache.get(pathToHandler)
        await proxyRequest(req, res, { hostname, protocol, port, secure })
        logger.debug(`Delivered ${path} by a cached handler`)
        return
      } catch (err) {
        logger.warn(`The handler for ${path} could not be used anymore.`)
        cache.delete(pathToHandler)
      }
    }

    // stage 2
    for (const frontend of frontends) {
      if (!frontend.path) continue
      const isRegExp = frontend.path instanceof RegExp
      const shouldHandle = isRegExp ? frontend.path.test(path) : path.startsWith(frontend.path)
      if (!shouldHandle) continue
      const { hostname, protocol, port } = frontend.target
      logger.info(`Proxy ${path} to frontend ${protocol}//${hostname}:${port}`)
      try {
        await proxyRequest(req, res, { hostname, protocol, port, secure: frontend.secure })
        cache.set(pathToHandler, frontend)
        logger.debug(`Found ${path} at ${protocol}//${hostname}:${port}`)
        return
      } catch (err) {
        logger.debug(`Did not find ${path} at ${protocol}//${hostname}:${port}, try next.`)
      }
    }

    // stage 3
    logger.debug(`Fall back to autofind ${path} in frontends`)
    const errors = []
    for (const frontend of frontends) {
      const { hostname, protocol, port } = frontend.target
      try {
        await proxyRequest(req, res, { hostname, protocol, port, secure: frontend.secure })
        cache.set(pathToHandler, frontend)
        logger.debug(`Found ${path} at ${protocol}//${hostname}:${port}`)
        return
      } catch (err) {
        logger.debug(`Did not find ${path} at ${protocol}//${hostname}:${port}, try next.`)
        errors.push(err)
      }
    }

    const err = errors.reduce((memo, err) => {
      // keep the first non-http error and forward it
      if (!(memo instanceof HttpError)) return memo
      // otherwise, use the last http error
      return err
    }, new createError.NotFound())

    const handle = finalhandler(req, res)
    if (err instanceof HttpError) {
      // in case of http-error the proxy explicitedly received a error code
      logger.warn(`Could not find ${path} in frontends`)
      return handle(err)
    }
    logger.error(`Something went wrong, while resolving a file in frontends: ${err.stack}`)
    return handle()
  }
}
